import { signal, computed, isDevMode, isSignal, DestroyRef, inject } from '@angular/core';
import { Subject, finalize, tap, EMPTY, catchError, first, of, switchMap, map, filter, merge, isObservable, defer, defaultIfEmpty, forkJoin, retry } from 'rxjs';
import { takeUntilDestroyed } from '@angular/core/rxjs-interop';

function isEmptyValue(variable) {
    if ((typeof variable === 'undefined') || variable === null || (variable !== variable)) {
        return true;
    }
    if (variable === 0 || variable === '0' || variable === 'false' || typeof variable === 'boolean') {
        return false;
    }
    if (typeof variable === 'object') {
        if (Array.isArray(variable)) {
            return variable.length === 0 || (variable.length === 1 && isEmptyObject(variable[0]));
        }
        return false;
    }
    return !variable;
}
function isEmptyObject(variable) {
    if (variable == null || isEmptyValue(variable)) {
        return true;
    }
    if (typeof variable === 'object') {
        try {
            const props = Object.keys(variable);
            if (!props || !props.length) {
                return true;
            }
        }
        catch (e) {
            console?.error(e);
        }
    }
    return false;
}
function getObjectPathValue(object, path, recLevel) {
    if (object == null) {
        return undefined;
    }
    if (!path.includes('.')) {
        return object[path];
    }
    try {
        const paths = path.split('.');
        const key = paths.shift();
        if (!key) {
            return undefined;
        }
        if (paths.length > 0) {
            const rec = recLevel || 0;
            if (rec > 100) {
                console.error('getObjectPathValue endless recursion');
                return undefined;
            }
            return getObjectPathValue(object[key], paths.join('.'), rec + 1);
        }
        else {
            return object[key];
        }
    }
    catch (e) {
        console.error(e);
        return undefined;
    }
}

const defaultComparatorFields = ['uuId', 'id'];

class Comparator {
    constructor(fields = defaultComparatorFields) {
        if (!fields) {
            this.renderDefaultComparatorError = true;
            this.fields = defaultComparatorFields;
        }
        else {
            this.fields = fields;
            this.renderDefaultComparatorError = false;
        }
    }
    equal(obj1, obj2, byFields) {
        if (Object.is(obj1, obj2)) {
            return true;
        }
        if (obj1 == null || obj1 !== obj1 || obj2 == null || obj2 !== obj2 || (typeof obj1 !== 'object') || (typeof obj2 !== 'object')) {
            return false;
        }
        const fields = byFields?.length ? byFields : this.fields;
        if (fields.length) {
            for (const field of fields) {
                if (typeof field === 'string') {
                    const value1 = getObjectPathValue(obj1, field);
                    const value2 = getObjectPathValue(obj2, field);
                    if (value1 !== undefined && value2 !== undefined) {
                        if (fieldValuesAreNotEmptyAndEqual(obj1, obj2, field)) {
                            return true;
                        }
                        else {
                            return false;
                        }
                    }
                }
                else {
                    if (!field.find(key => !fieldValuesAreNotEmptyAndEqual(obj1, obj2, key))) {
                        return true;
                    }
                }
            }
        }
        return false;
    }
}
function fieldValuesAreNotEmptyAndEqual(obj1, obj2, field) {
    const value1 = getObjectPathValue(obj1, field);
    if (value1 === undefined) {
        return false;
    }
    const value2 = getObjectPathValue(obj2, field);
    if (value2 === undefined) {
        return false;
    }
    return (!isEmptyValue(value1)
        && !isEmptyValue(value2)
        && !isEmptyObject(value1)
        && !isEmptyObject(value2)
        && value1 === value2);
}
class DuplicateError extends Error {
    constructor(...params) {
        super(...params);
        this.name = 'DuplicateError';
        if (Error.captureStackTrace) {
            Error.captureStackTrace(this, DuplicateError);
        }
    }
}

class Collection {
    constructor(options) {
        this.comparator = new Comparator();
        this.allowFetchedDuplicates = true;
        this.onDuplicateErrCallbackParam = new DuplicateError();
        this.onCreate = new Subject();
        this.onRead = new Subject();
        this.onUpdate = new Subject();
        this.onDelete = new Subject();
        this.isFirstItemsRequest = true;
        this.equalItems = (a, b) => {
            if (!a || !b || a == null || b == null || !Array.isArray(a) || !Array.isArray(b) || a.length !== b.length) {
                return false;
            }
            for (let i = 0; i < a.length; i++) {
                if (!this.comparator.equal(a[i], b[i])) {
                    return false;
                }
            }
            return true;
        };
        this.state = {
            $items: signal([]),
            $totalCountFetched: signal(undefined),
            $isCreating: signal(false),
            $isReading: signal(false),
            $isBeforeFirstRead: signal(true),
            $updatingItems: signal([], { equal: this.equalItems }),
            $deletingItems: signal([], { equal: this.equalItems }),
            $refreshingItems: signal([], { equal: this.equalItems }),
            $status: signal(new Map()),
            $statuses: signal(new Map()),
        };
        this.$items = computed(() => {
            if (this.isFirstItemsRequest) {
                this.isFirstItemsRequest = false;
                if (this.onFirstItemsRequest) {
                    Promise.resolve().then(() => {
                        try {
                            this.onFirstItemsRequest?.();
                        }
                        catch (e) {
                            this.reportRuntimeError(e);
                        }
                    });
                }
            }
            return this.state.$items();
        });
        this.$totalCountFetched = this.state.$totalCountFetched.asReadonly();
        this.$isCreating = this.state.$isCreating.asReadonly();
        this.$isReading = this.state.$isReading.asReadonly();
        this.$isBeforeFirstRead = this.state.$isBeforeFirstRead.asReadonly();
        this.$updatingItems = computed(() => {
            const updating = this.state.$updatingItems();
            const existing = this.$items();
            return updating.filter(item => this.hasItemIn(item, existing));
        });
        this.$deletingItems = computed(() => {
            const deleting = this.state.$deletingItems();
            const existing = this.$items();
            return deleting.filter(item => this.hasItemIn(item, existing));
        });
        this.$refreshingItems = computed(() => {
            const refreshing = this.state.$refreshingItems();
            const existing = this.$items();
            return refreshing.filter(item => this.hasItemIn(item, existing));
        });
        this.$status = this.state.$status.asReadonly();
        this.$statuses = this.state.$statuses.asReadonly();
        this.$isUpdating = computed(() => this.state.$updatingItems().length > 0);
        this.$isDeleting = computed(() => this.state.$deletingItems().length > 0);
        this.$isMutating = computed(() => {
            const isCreating = this.state.$isCreating();
            const isUpdating = this.$isUpdating();
            const isDeleting = this.$isDeleting();
            return isCreating || isUpdating || isDeleting;
        });
        this.$isSaving = computed(() => {
            const isCreating = this.state.$isCreating();
            const isUpdating = this.$isUpdating();
            return isCreating || isUpdating;
        });
        this.$isProcessing = computed(() => {
            const isMutating = this.$isMutating();
            const isReading = this.state.$isReading();
            const isRefreshing = this.state.$refreshingItems().length > 0;
            return isMutating || isReading || isRefreshing;
        });
        this.$mutatingItems = computed(() => this.uniqueItems([
            ...this.$updatingItems(),
            ...this.$deletingItems()
        ]));
        this.$processingItems = computed(() => this.uniqueItems([
            ...this.$mutatingItems(),
            ...this.$refreshingItems()
        ]));
        this.setOptions(options);
        this.init();
        Promise.resolve().then(() => {
            this.asyncInit();
            if (isDevMode() && this.comparator.hasOwnProperty('renderDefaultComparatorError') && this.comparator.renderDefaultComparatorError === true) {
                this.reportRuntimeError('Default Comparator with default id fields will be used.\n' +
                    'Don\'t forget to set option \'comparatorFields\' or \'comparator\' using constructor(), setComparator(), init(), or asyncInit().', 'Collection:', this.constructor.name, 'Comparator fields:', defaultComparatorFields);
            }
        });
    }
    init() {
    }
    asyncInit() {
    }
    create(params) {
        this.state.$isCreating.set(true);
        return this.toObservableFirstValue(params.request).pipe(finalize(() => this.state.$isCreating.set(false)), tap((item) => {
            if (item != null) {
                this.state.$items.update((items) => {
                    if (!this.hasItemIn(item, items)) {
                        this.callCb(params.onSuccess, item);
                        if (this.onCreate.observed) {
                            this.onCreate.next([item]);
                        }
                        return [...items, item];
                    }
                    else {
                        this.duplicateNotAdded(item, items);
                        this.callCb(params.onError, this.onDuplicateErrCallbackParam);
                        return items;
                    }
                });
            }
        }), catchError((error) => {
            this.callCb(params.onError, error);
            return EMPTY;
        }));
    }
    createMany(params) {
        this.state.$isCreating.set(true);
        const request = Array.isArray(params.request) ?
            this.forkJoinSafe(params.request) :
            this.toObservableFirstValue(params.request);
        return request.pipe(finalize(() => this.state.$isCreating.set(false)), tap((fetched) => {
            if (fetched != null) {
                this.state.$totalCountFetched.set(!Array.isArray(fetched) ? fetched.totalCount : undefined);
                const fetchedItems = Array.isArray(fetched) ? fetched : fetched.items;
                if (fetchedItems.length > 0) {
                    this.state.$items.update((items) => {
                        const nextItems = items.slice();
                        const result = this.upsertMany(fetchedItems, nextItems);
                        if (!Array.isArray(result) && result.duplicate) {
                            this.duplicateNotAdded(result.duplicate, result.preExisting ? fetchedItems : items);
                            this.callCb(params.onError, this.onDuplicateErrCallbackParam);
                            return items;
                        }
                        else {
                            this.callCb(params.onSuccess, fetchedItems);
                            if (this.onCreate.observed) {
                                this.onCreate.next(fetchedItems);
                            }
                            return nextItems;
                        }
                    });
                }
            }
        }), catchError((error) => {
            this.callCb(params.onError, error);
            return EMPTY;
        }));
    }
    read(params) {
        this.state.$isReading.set(true);
        return this.toObservableFirstValue(params.request).pipe(finalize(() => {
            this.state.$isReading.set(false);
            this.state.$isBeforeFirstRead.set(false);
        }), tap((fetched) => {
            const items = fetched == null ? [] : (Array.isArray(fetched) ? fetched : fetched.items);
            if (items.length > 0) {
                const duplicate = this.hasDuplicates(items);
                if (duplicate != null) {
                    this.duplicateNotAdded(duplicate, items);
                    if (!this.allowFetchedDuplicates) {
                        if (!params.keepExistingOnError) {
                            this.state.$items.set([]);
                            this.state.$totalCountFetched.set(undefined);
                        }
                        this.callCb(params.onError, this.onDuplicateErrCallbackParam);
                        return;
                    }
                }
            }
            this.state.$items.set(items);
            this.state.$totalCountFetched.set(Array.isArray(fetched) ? undefined : fetched.totalCount);
            this.callCb(params.onSuccess, items);
            if (this.onRead.observed) {
                this.onRead.next(items);
            }
        }), catchError((error) => {
            if (!params.keepExistingOnError) {
                this.state.$items.set([]);
                this.state.$totalCountFetched.set(undefined);
            }
            this.callCb(params.onError, error);
            return EMPTY;
        }));
    }
    readOne(params) {
        this.state.$isReading.set(true);
        return this.toObservableFirstValue(params.request).pipe(finalize(() => {
            this.state.$isReading.set(false);
            this.state.$isBeforeFirstRead.set(false);
        }), tap((newItem) => {
            if (newItem != null) {
                this.state.$items.update((items) => {
                    const nextItems = items.slice();
                    if (this.upsertOne(newItem, nextItems) === 'duplicate') {
                        this.duplicateNotAdded(newItem, items);
                        this.callCb(params.onError, this.onDuplicateErrCallbackParam);
                        return items;
                    }
                    else {
                        this.callCb(params.onSuccess, newItem);
                        if (this.onRead.observed) {
                            this.onRead.next([newItem]);
                        }
                        return nextItems;
                    }
                });
            }
        }), catchError((error) => {
            this.callCb(params.onError, error);
            return EMPTY;
        }));
    }
    readMany(params) {
        this.state.$isReading.set(true);
        const request = Array.isArray(params.request) ? this.forkJoinSafe(params.request) : this.toObservableFirstValue(params.request);
        return request.pipe(finalize(() => {
            this.state.$isReading.set(false);
            this.state.$isBeforeFirstRead.set(false);
        }), tap((fetched) => {
            this.state.$totalCountFetched.set(!Array.isArray(fetched) ? fetched.totalCount : undefined);
            const readItems = fetched == null ? [] : (Array.isArray(fetched) ? fetched : fetched.items);
            if (readItems.length > 0) {
                this.state.$items.update((items) => {
                    const nextItems = items.slice();
                    const result = this.upsertMany(readItems, nextItems);
                    if (!Array.isArray(result) && result.duplicate) {
                        this.duplicateNotAdded(result.duplicate, result.preExisting ? readItems : items);
                        this.callCb(params.onError, this.onDuplicateErrCallbackParam);
                        return items;
                    }
                    else {
                        this.callCb(params.onSuccess, nextItems);
                        if (this.onRead.observed) {
                            this.onRead.next(readItems);
                        }
                        return nextItems;
                    }
                });
            }
        }), catchError((error) => {
            this.callCb(params.onError, error);
            return EMPTY;
        }));
    }
    refresh(params) {
        this.state.$refreshingItems.update((refreshingItems) => this.getWith(refreshingItems, params.item));
        return this.toObservableFirstValue(params.request).pipe(finalize(() => this.state.$refreshingItems.update((refreshingItems) => this.getWithout(refreshingItems, params.item))), tap((newItem) => {
            if (newItem != null) {
                this.state.$items.update((items) => {
                    const nextItems = items.slice();
                    if (this.upsertOne(newItem, nextItems) === 'duplicate') {
                        this.duplicateNotAdded(newItem, items);
                        this.callCb(params.onError, this.onDuplicateErrCallbackParam);
                        return items;
                    }
                    else {
                        this.callCb(params.onSuccess, newItem);
                        if (this.onRead.observed) {
                            this.onRead.next([newItem]);
                        }
                        return nextItems;
                    }
                });
            }
        }), catchError((error) => {
            this.callCb(params.onError, error);
            return EMPTY;
        }));
    }
    refreshMany(params) {
        this.state.$refreshingItems.update((refreshingItems) => this.getWith(refreshingItems, params.items));
        const request = Array.isArray(params.request) ? this.forkJoinSafe(params.request) : this.toObservableFirstValue(params.request);
        return request.pipe(finalize(() => this.state.$refreshingItems.update((refreshingItems) => this.getWithout(refreshingItems, params.items))), tap((fetched) => {
            if (fetched != null) {
                this.state.$totalCountFetched.set(!Array.isArray(fetched) ? fetched.totalCount : undefined);
                const fetchedItems = Array.isArray(fetched) ? fetched : fetched.items;
                if (fetchedItems.length > 0) {
                    this.state.$items.update((items) => {
                        const nextItems = items.slice();
                        const result = this.upsertMany(fetchedItems, nextItems);
                        if (!Array.isArray(result) && result.duplicate) {
                            this.duplicateNotAdded(result.duplicate, result.preExisting ? fetchedItems : items);
                            this.callCb(params.onError, this.onDuplicateErrCallbackParam);
                            return items;
                        }
                        else {
                            this.callCb(params.onSuccess, nextItems);
                            if (this.onRead.observed) {
                                this.onRead.next(fetchedItems);
                            }
                            return nextItems;
                        }
                    });
                }
            }
        }), catchError((error) => {
            this.callCb(params.onError, error);
            return EMPTY;
        }));
    }
    update(params) {
        this.state.$updatingItems.update((updatingItems) => this.getWith(updatingItems, params.item));
        return this.toObservableFirstValue(params.request).pipe(finalize(() => this.state.$updatingItems.update((updatingItems) => this.getWithout(updatingItems, params.item))), first(), switchMap((newItem) => {
            if (params.refreshRequest) {
                return this.refresh({
                    request: params.refreshRequest,
                    item: params.item,
                    onSuccess: params.onSuccess,
                    onError: params.onError,
                });
            }
            else {
                if (newItem != null) {
                    this.state.$items.update((items) => {
                        const nextItems = items.slice();
                        if (this.upsertOne(newItem, nextItems) === 'duplicate') {
                            this.duplicateNotAdded(newItem, items);
                            this.callCb(params.onError, this.onDuplicateErrCallbackParam);
                            return items;
                        }
                        else {
                            this.callCb(params.onSuccess, newItem);
                            if (this.onUpdate.observed) {
                                this.onUpdate.next([newItem]);
                            }
                            return nextItems;
                        }
                    });
                }
                return of(newItem);
            }
        }), catchError((error) => {
            this.callCb(params.onError, error);
            return EMPTY;
        }));
    }
    updateMany(params) {
        this.state.$updatingItems.update((updatingItems) => this.getWith(updatingItems, params.items));
        const request = Array.isArray(params.request) ? this.forkJoinSafe(params.request) : this.toObservableFirstValue(params.request);
        return request.pipe(finalize(() => this.state.$updatingItems.update((updatingItems) => this.getWithout(updatingItems, params.items))), first(), switchMap((updatedItems) => {
            if (params.refreshRequest) {
                return this.refreshMany({
                    request: params.refreshRequest,
                    items: params.items,
                    onSuccess: params.onSuccess,
                    onError: params.onError,
                });
            }
            else {
                if (updatedItems != null) {
                    this.state.$items.update((items) => {
                        const nextItems = items.slice();
                        const result = this.upsertMany(updatedItems, nextItems);
                        if (!Array.isArray(result) && result.duplicate) {
                            this.duplicateNotAdded(result.duplicate, result.preExisting ? updatedItems : items);
                            this.callCb(params.onError, this.onDuplicateErrCallbackParam);
                            return items;
                        }
                        else {
                            this.callCb(params.onSuccess, updatedItems);
                            if (this.onUpdate.observed) {
                                this.onUpdate.next(updatedItems);
                            }
                            return nextItems;
                        }
                    });
                }
                return of(updatedItems);
            }
        }), catchError((error) => {
            this.callCb(params.onError, error);
            return EMPTY;
        }));
    }
    delete(params) {
        this.state.$deletingItems.update((deletingItems) => this.getWith(deletingItems, params.item));
        const req = params.request ? this.toObservableFirstValue(params.request) : of(null);
        return req.pipe(finalize(() => this.state.$deletingItems.update((deletingItems) => this.getWithout(deletingItems, params.item))), switchMap((response) => {
            if (params.readRequest) {
                return this.read({
                    request: params.readRequest,
                    onSuccess: params.onSuccess ? () => params.onSuccess?.(response) : undefined,
                    onError: params.onError,
                }).pipe(map(_ => response));
            }
            else {
                this.state.$totalCountFetched.update((prevTotalCount) => this.getDecrementedTotalCount([response], 1, prevTotalCount, params.decrementTotalCount) ?? prevTotalCount);
                this.state.$items.update((items) => items.filter(item => !this.comparator.equal(item, params.item)));
                this.callCb(params.onSuccess, response);
                if (this.onDelete.observed) {
                    this.onDelete.next([params.item]);
                }
                return of(response);
            }
        }), catchError((error) => {
            this.callCb(params.onError, error);
            return EMPTY;
        }));
    }
    deleteMany(params) {
        this.state.$deletingItems.update((deletingItems) => this.getWith(deletingItems, params.items));
        const requests = Array.isArray(params.request) ?
            this.forkJoinSafe(params.request)
            : this.forkJoinSafe([params.request ?? of(null)]);
        return requests.pipe(finalize(() => this.state.$deletingItems.update((deletingItems) => this.getWithout(deletingItems, params.items))), first(), switchMap((response) => {
            if (params.readRequest) {
                return this.read({
                    request: params.readRequest,
                    onSuccess: params.onSuccess ? () => params.onSuccess?.(response) : undefined,
                    onError: params.onError,
                }).pipe(map(_ => response));
            }
            else {
                this.state.$totalCountFetched.update((prevTotalCount) => this.getDecrementedTotalCount(response, params.items.length, prevTotalCount, params.decrementTotalCount) ?? prevTotalCount);
                this.state.$items.update((items) => items.filter(item => !this.hasItemIn(item, params.items)));
                this.callCb(params.onSuccess, response);
                if (this.onDelete.observed) {
                    this.onDelete.next(params.items);
                }
                return of(response);
            }
        }), catchError((error) => {
            this.callCb(params.onError, error);
            return EMPTY;
        }));
    }
    setUniqueStatus(status, item, active = true) {
        const map = this.state.$status();
        const currentItem = map.get(status);
        if (!active) {
            if (currentItem == null || !this.comparator.equal(item, currentItem)) {
                return;
            }
            else {
                const newMap = new Map(map);
                newMap.delete(status);
                this.state.$status.set(newMap);
            }
        }
        else {
            if (currentItem != null && this.comparator.equal(item, currentItem)) {
                return;
            }
            else {
                const newMap = new Map(map);
                newMap.set(status, item);
                this.state.$status.set(newMap);
            }
        }
    }
    deleteUniqueStatus(status) {
        const map = this.state.$status();
        if (map.has(status)) {
            const newMap = new Map(map);
            newMap.delete(status);
            this.state.$status.set(newMap);
        }
    }
    setItemStatus(item, status) {
        const map = this.state.$statuses();
        if (map.get(item)?.has(status)) {
            return;
        }
        const newMap = new Map(map);
        const itemStatuses = newMap.get(item) ?? new Set();
        itemStatuses.add(status);
        newMap.set(item, itemStatuses);
        this.state.$statuses.set(newMap);
    }
    deleteItemStatus(item, status) {
        const map = this.state.$statuses();
        const current = map.get(item);
        if (current && !current.has(status)) {
            return;
        }
        const newMap = new Map(map);
        const itemStatuses = newMap.get(item);
        if (itemStatuses) {
            itemStatuses.delete(status);
            newMap.set(item, itemStatuses);
            this.state.$statuses.set(newMap);
        }
    }
    isItemDeleting(itemSource) {
        return computed(() => {
            const i = isSignal(itemSource) ? itemSource() : itemSource;
            return !!i && this.hasItemIn(i, this.state.$deletingItems());
        });
    }
    isItemRefreshing(itemSource) {
        return computed(() => {
            const i = isSignal(itemSource) ? itemSource() : itemSource;
            return !!i && this.hasItemIn(i, this.state.$refreshingItems());
        });
    }
    isItemUpdating(itemSource) {
        return computed(() => {
            const i = isSignal(itemSource) ? itemSource() : itemSource;
            return !!i && this.hasItemIn(i, this.state.$updatingItems());
        });
    }
    isItemMutating(itemSource) {
        return computed(() => {
            const i = isSignal(itemSource) ? itemSource() : itemSource;
            return !!i && this.hasItemIn(i, this.$mutatingItems());
        });
    }
    isItemProcessing(itemSource) {
        return computed(() => {
            const i = isSignal(itemSource) ? itemSource() : itemSource;
            return !!i && this.$isProcessing() && (this.hasItemIn(i, this.state.$refreshingItems()) || this.hasItemIn(i, this.$mutatingItems()));
        });
    }
    getItem(filter, equalFn) {
        if (isSignal(filter)) {
            return computed(() => {
                const item = filter();
                return item ? this.getItemByPartial(item, this.$items()) : undefined;
            }, { equal: equalFn });
        }
        else {
            return computed(() => filter ? this.getItemByPartial(filter, this.$items()) : undefined, {
                equal: equalFn
            });
        }
    }
    getItemByField(field, fieldValue, equalFn) {
        const fields = Array.isArray(field) ? field : [field];
        if (isSignal(fieldValue)) {
            return computed(() => {
                const fv = fieldValue();
                if (fv == null) {
                    return undefined;
                }
                return this.$items().find((i) => fields.find((f) => i[f] === fv) !== undefined);
            }, { equal: equalFn });
        }
        else {
            return computed(() => this.$items().find((i) => fields.find((f) => i[f] === fieldValue) !== undefined), { equal: equalFn });
        }
    }
    setThrowOnDuplicates(message) {
        this.throwOnDuplicates = message;
    }
    setAllowFetchedDuplicates(allowFetchedDuplicates) {
        this.allowFetchedDuplicates = allowFetchedDuplicates;
    }
    setOnDuplicateErrCallbackParam(onDuplicateErrCallbackParam) {
        this.onDuplicateErrCallbackParam = onDuplicateErrCallbackParam;
    }
    hasItemIn(item, arr) {
        if (!arr || !Array.isArray(arr) || arr.length === 0) {
            return false;
        }
        return ((arr.findIndex(i => this.comparator.equal(i, item))) > -1);
    }
    listenForItemsUpdate(items) {
        return merge(this.onUpdate, this.onRead).pipe(map((updated) => this.findInSet(items, updated)), filter(found => found.length > 0));
    }
    listenForItemsDeletion(items) {
        return this.onDelete.pipe(map((updated) => this.findInSet(items, updated)), filter(found => found.length > 0));
    }
    listenForCreate() {
        return this.onCreate.asObservable();
    }
    listenForRead() {
        return this.onRead.asObservable();
    }
    listenForUpdate() {
        return this.onUpdate.asObservable();
    }
    listenForDelete() {
        return this.onDelete.asObservable();
    }
    setComparator(comparator) {
        if (typeof comparator === 'function') {
            this.comparator = { equal: comparator };
        }
        else {
            this.comparator = comparator;
        }
    }
    setOptions(options) {
        if (options != null) {
            if (options.comparator) {
                this.setComparator(options.comparator);
            }
            else {
                if (options.comparatorFields != null) {
                    this.comparator = new Comparator(options.comparatorFields);
                }
            }
            if (options.throwOnDuplicates != null) {
                this.throwOnDuplicates = options.throwOnDuplicates;
            }
            if (typeof options.allowFetchedDuplicates === 'boolean') {
                this.allowFetchedDuplicates = options.allowFetchedDuplicates;
            }
            if (options.onDuplicateErrCallbackParam != null) {
                this.onDuplicateErrCallbackParam = options.onDuplicateErrCallbackParam;
            }
            if (options.errReporter != null && typeof options.errReporter === 'function') {
                this.errReporter = options.errReporter;
            }
            if (options.onFirstItemsRequest) {
                this.onFirstItemsRequest = options.onFirstItemsRequest;
            }
        }
    }
    uniqueItems(items) {
        if (!Array.isArray(items)) {
            throw new Error('uniqueItems expects an array');
        }
        return items.filter((item, index) => items.indexOf(item) === index
            || items.findIndex(i => this.comparator.equal(i, item)) == index);
    }
    getItemByPartial(partItem, items) {
        if (!items) {
            items = this.$items();
        }
        return items.find(i => this.comparator.equal(i, partItem));
    }
    getDuplicates(items) {
        if (!Array.isArray(items) || items.length < 2) {
            return null;
        }
        const duplicates = {};
        let hasDupes = false;
        for (let ii = 0; ii < items.length; ii++) {
            const iDupes = {};
            let itemHasDupes = false;
            const i = items[ii];
            for (let ij = ii + 1; ij < items.length; ij++) {
                const j = items[ij];
                if (this.comparator.equal(i, j)) {
                    iDupes[ij] = j;
                    itemHasDupes = true;
                }
            }
            if (itemHasDupes) {
                hasDupes = true;
                iDupes[ii] = i;
                duplicates[ii] = iDupes;
            }
        }
        return hasDupes ? duplicates : null;
    }
    setOnFirstItemsRequest(cb) {
        this.onFirstItemsRequest = cb;
    }
    getWith(items, item) {
        const addItems = Array.isArray(item) ?
            item.map(i => this.getItemByPartial(i, items) ?? i)
            : [this.getItemByPartial(item, items) ?? item];
        return this.uniqueItems([
            ...items,
            ...addItems
        ]);
    }
    getWithout(items, item) {
        if (Array.isArray(item)) {
            return items.filter(i => !this.hasItemIn(i, item));
        }
        else {
            return items.filter(i => !this.comparator.equal(i, item));
        }
    }
    findInSet(seek, items) {
        return items.filter(i => this.hasItemIn(i, seek));
    }
    upsertOne(item, items) {
        let firstIndex = -1;
        let duplicateIndex = -1;
        for (let i = 0; i < items.length; i++) {
            if (this.comparator.equal(item, items[i])) {
                if (firstIndex < 0) {
                    firstIndex = i;
                }
                else {
                    duplicateIndex = i;
                    break;
                }
            }
        }
        if (firstIndex < 0) {
            items.push(item);
        }
        else {
            if (duplicateIndex < 0) {
                items[firstIndex] = item;
            }
            else {
                return 'duplicate';
            }
        }
        return items;
    }
    hasDuplicates(items) {
        if (!Array.isArray(items) || items.length < 2) {
            return null;
        }
        for (let i = 0; i < items.length; i++) {
            for (let j = i + 1; j < items.length; j++) {
                if (this.comparator.equal(items[i], items[j])) {
                    return items[j];
                }
            }
        }
        return null;
    }
    upsertMany(items, toItems) {
        const hasDuplicates = this.hasDuplicates(items);
        if (hasDuplicates) {
            return { duplicate: hasDuplicates };
        }
        for (let i = 0; i < items.length; i++) {
            if (this.upsertOne(items[i], toItems) === 'duplicate') {
                return { duplicate: items[i] };
            }
        }
        return toItems;
    }
    duplicateNotAdded(item, items) {
        if (this.throwOnDuplicates) {
            throw new Error(this.throwOnDuplicates);
        }
        if (this.errReporter) {
            this.errReporter('Duplicate can not be added to collection:', item, items);
            const duplicates = this.getDuplicates(items);
            if (duplicates) {
                this.errReporter('Duplicates are found in collection:', duplicates);
            }
        }
    }
    callCb(cb, param) {
        if (cb != null) {
            try {
                cb(param);
            }
            catch (e) {
                this.reportRuntimeError(e);
            }
        }
    }
    callErrReporter(...args) {
        if (this.errReporter) {
            try {
                this.errReporter(...args);
            }
            catch (e) {
                console?.error('errReporter threw an exception', e);
            }
        }
    }
    reportRuntimeError(...args) {
        if (this.errReporter) {
            this.callErrReporter(...args);
        }
        else {
            console?.error(...args);
        }
    }
    getDecrementedTotalCount(response, itemsCount, prevTotalCount, decrementTotalCount) {
        try {
            if (decrementTotalCount != null) {
                if (typeof decrementTotalCount === 'string') {
                    if (response
                        && Array.isArray(response)
                        && response.length === 1
                        && typeof response[0] === 'object'
                        && response[0] != null) {
                        const r = response[0];
                        if (Object.hasOwn(r, decrementTotalCount)
                            && typeof r[decrementTotalCount] === 'number'
                            && r[decrementTotalCount] >= 0
                            && Math.round(r[decrementTotalCount]) === r[decrementTotalCount]) {
                            return r[decrementTotalCount];
                        }
                    }
                }
                else {
                    if (prevTotalCount) {
                        if (typeof decrementTotalCount === 'number') {
                            const decr = Math.round(decrementTotalCount);
                            if (decr === decrementTotalCount && decr > 0 && prevTotalCount >= decr) {
                                return prevTotalCount - decrementTotalCount;
                            }
                        }
                        if (typeof decrementTotalCount === 'boolean') {
                            if (prevTotalCount >= itemsCount) {
                                return prevTotalCount - itemsCount;
                            }
                        }
                    }
                }
            }
        }
        catch (e) {
            this.reportRuntimeError(e);
        }
        return null;
    }
    toObservableFirstValue(s) {
        return isObservable(s) ? s.pipe(first()) : defer(() => of(s()));
    }
    forkJoinSafe(sources) {
        const source = sources.map(s => this.toObservableFirstValue(s).pipe(map(v => ({ value: v, error: false })), defaultIfEmpty({ error: true, value: undefined }), catchError((e) => {
            this.callErrReporter(e);
            return of({ error: true, value: undefined });
        })));
        return forkJoin(source).pipe(map((values) => {
            return values.filter(v => !v.error).map(v => v.value);
        }));
    }
}

function createEffect(generator) {
    const destroyRef = inject(DestroyRef);
    const origin$ = new Subject();
    generator(origin$).pipe(retry(), takeUntilDestroyed(destroyRef)).subscribe();
    return ((observableOrValue) => {
        const observable$ = isObservable(observableOrValue)
            ? observableOrValue.pipe(retry())
            : of(observableOrValue);
        return observable$.pipe(takeUntilDestroyed(destroyRef)).subscribe((value) => {
            origin$.next(value);
        });
    });
}

export { Collection, Comparator, DuplicateError, createEffect, createEffect as effect, createEffect as sideEffect };
//# sourceMappingURL=ngx-collection.mjs.map
